<?php

// NOTE: See HowToImport.txt for details.
//
// Version: 1.0
// Date: 11/08/2009
// Author: Valentin Guignon (valentin.guignon@cirad.fr)
// License: GNU GENERAL PUBLIC LICENSE

/*****************************************************************************/
// Global config variables
//
// import_user_names: if set to true, the importer tries to import user names
//    using the 'import_user_names' associative array first and then see if
//    Drupal has a user name corresponding to the last contributor of the
//    imported wiki page.
$GLOBALS['import_user_names'] = TRUE;
// user_name_lookup: used to give a Drupal user the ownership of the wiki
//    content if he/she was the last contributor
$GLOBALS['user_name_lookup'] = array(
//    format:
//        "MediaWiki contributor" => "Drupal user login",
    );

// node_type: name of the node type you created under Drupal for wiki content
$GLOBALS['node_type'] = 'mediawiki';
// mediawiki_input_format_number: correspond to the number of the "Wiki" input
//    format. See on the creation form of a Wiki content; usually 1 is
//    "Filtered HTML", 2 is "Full HTML", 3 is "PHP code" and 4 is "Wiki".
//+FIXME: auto-guess the number?
$GLOBALS['mediawiki_input_format_number'] = 4;

// convert_mediawiki_tags_to_html: if set to false, wiki tags are kept
//    otherwise they are turned into html tags which may be better for
//    later editing with Drupal.
$GLOBALS['convert_mediawiki_tags_to_html'] = TRUE;

// NOTE: You may also have to adjust the variables "$url_to_files" and
// "$url_to_drupal_wiki" in the convertMediaWikiToHTML function.


// Global variables (do not edit)
$pages_count          = 0; // processed pages in XML
$imported_pages_count = 0; // imported pages

// used during XML parsing...
$inside_page   = FALSE;
$inside_title  = FALSE;
$inside_text   = FALSE;
$inside_author = FALSE;

$current_title   = '';
$current_content = '';
$current_author  = '';
/*****************************************************************************/




/**
* convertMediaWikiToHTML
*************************
Converts a MediaWiki content into HTML.
For instance "== Title 2 ==" is turned into "<h2>Title 2</h2>".

Parameters:
 $text (string): the MediaWiki content to convert.

Return:
 (string) the HTML content
*/
function convertMediaWikiToHTML($text)
{

/************************************/
/*** Mediawiki input filter START ***/
/************************************/

// url where you put the MediaWiki files;
//    ex.: "http://www.myserver.com/drupal/sites/default/files/mediawiki/"
$url_to_files = url('', array('absolute' => TRUE))
                . 'sites/default/files/mediawiki/';
// url used by Wikitools; ex.: "http://www.myserver.com/drupal/?q=wiki/"
$url_to_drupal_wiki = url('wiki/');

$patterns = array();
$replacements = array();

// add EOL to the last line if it contains a Wiki tag
array_push($patterns, '/([\n\r](?: |&nbsp;)*[*#:|][^\r\n]+)$/s');
array_push($replacements, "$1\n");

// = h1 =
array_push($patterns, '/^(?: |&nbsp;)*=([^=\n\r][^\n\r]*?)=(?: |&nbsp;)*/m');
array_push($replacements, "<h1>$1</h1>");
//+FIXME: add anchors (with _valid_ names)
//array_push($replacements, "<a name=\"$1\"></a>\n<h1>$1</h1>");

// == h2 ==
array_push($patterns, '/^(?: |&nbsp;)*==([^=\n\r][^\n\r]*?)==(?: |&nbsp;)*/m');
array_push($replacements, "<h2>$1</h2>");

// === h3 ===
array_push($patterns, '/^(?: |&nbsp;)*===([^=\n\r][^\n\r]*?)===(?: |&nbsp;)*/m');
array_push($replacements, "<h3>$1</h3>");

// ==== h4 ====
array_push($patterns, '/^(?: |&nbsp;)*====([^=\n\r][^\n\r]*?)====(?: |&nbsp;)*/m');
array_push($replacements, "<h4>$1</h4>");

// '''bold'''
array_push($patterns, '/\'\'\'(.+?)\'\'\'/');
array_push($replacements, '<b>$1</b>');

// ''italic''
array_push($patterns, '/\'\'(.+?)\'\'/');
array_push($replacements, '<i>$1</i>');

// [[image:file|thumb|alt text]]
array_push($patterns, '/\[\[Image:([^\|\]"\n\r]+)\|thumb\|([^\]\n\r]*)\]\]/i');
array_push($replacements, "<img class=\"mediawiki_thumb\" src=\"$url_to_files$1\" alt=\"$2\"/>");

// [[image:file|alt text]]
array_push($patterns, '/\[\[Image:([^\|\]"\n\r]+\.(?:jpg|jpeg|png|gif|bmp|ico))\|?([^\]\n\r]*)\]\]/i');
array_push($replacements, "<img src=\"$url_to_files$1\" alt=\"$2\"/>");

// [[image:file]]
array_push($patterns, '/\[\[Image:([^\|\]"\n\r]+\.(?:jpg|jpeg|png|gif|bmp|ico))\]\]/i');
array_push($replacements, "<img src=\"$url_to_files$1\" alt=\"$1\"/>");

// no-image files:
// [[image:file|alt text]]
array_push($patterns, '/\[\[Image:([^\|\]"\n\r]+)\|?([^\]\n\r]*)\]\]/i');
array_push($replacements, "<a href=\"$url_to_files$1\">$2</a>");

// [[image:file]]
array_push($patterns, '/\[\[Image:([^\|\]"\n\r]+)\]\]/i');
array_push($replacements, "<a href=\"$url_to_files$1\">$1</a>");

// [[media:file]]
array_push($patterns, '/\[\[Media:([^\|\]"\n\r]+)\]\]/i');
array_push($replacements, "<a href=\"$url_to_files$1\">$1</a>");

// [[wiki link|Specified Name]]
array_push($patterns, '/\[\[([^\|\]"\n\r]+)\|([^\]\n\r]+)\]\]/');
array_push($replacements, "<a href=\"$url_to_drupal_wiki$1\">$2</a>");

// [[wiki link]]
array_push($patterns, '/\[\[([^\]"\n\r]+)\]\]/');
array_push($replacements, "<a href=\"$url_to_drupal_wiki$1\">$1</a>");

// [http://www.url.com/ Named URL]
array_push($patterns, '/\[(http|https|ftp):\/\/([^\]]+) ([^\]]+)]/');
array_push($replacements, '<a href="$1://$2">$3</a>');

// [http://www.url.com/]
array_push($patterns, '/\[(http|https|ftp):\/\/([^\]]+)]/');
array_push($replacements, '<a href="$1://$2">$1://$2</a>');


//+FIXME: lists (starting with *, # or :) with sub-items not taken in account
// * list of elements
array_push($patterns, '/^((?:(?: |&nbsp;)*\*[^\r\n][^\r\n]*[\r\n]{1,2})+)/ms');
array_push($replacements, "<ul class=\"mediawiki\">\n$1</ul>\n");
array_push($patterns, '/^(?: |&nbsp;)*\*([^\*\r\n].*)$/m');
array_push($replacements, "<li>$1</li>");
// second level
array_push($patterns, '/^((?:(?: |&nbsp;)*\*[^\r\n][^\r\n]*[\r\n]{1,2})+)/ms');
array_push($replacements, "<li style=\"background-image: none;\">\n<ul class=\"mediawiki\">\n$1</ul>\n</li>\n");
array_push($patterns, '/^(?: |&nbsp;)*\*\*([^\*\r\n].*)$/m');
array_push($replacements, "<li>$1</li>");
// third level
array_push($patterns, '/^((?:(?: |&nbsp;)*\*[^\r\n][^\r\n]*[\r\n]{1,2})+)/ms');
array_push($replacements, "<li style=\"background-image: none;\">\n<ul class=\"mediawiki\">\n$1</ul>\n</li>\n");
array_push($patterns, '/^(?: |&nbsp;)*\*\*\*([^\*\r\n].*)$/m');
array_push($replacements, "<li>$1</li>");
// fourth level
array_push($patterns, '/^((?:(?: |&nbsp;)*\*[^\r\n][^\r\n]*[\r\n]{1,2})+)/ms');
array_push($replacements, "<li style=\"background-image: none;\">\n<ul class=\"mediawiki\">\n$1</ul>\n</li>\n");
array_push($patterns, '/^(?: |&nbsp;)*\*\*\*\*([^\*\r\n].*)$/m');
array_push($replacements, "<li>$1</li>");

// # list of elements
array_push($patterns, '/^((?:(?: |&nbsp;)*#[^\r\n][^\r\n]*[\r\n]{1,2})+)/ms');
array_push($replacements, "<ol class=\"mediawiki\">\n$1</ol>\n");
array_push($patterns, '/^(?: |&nbsp;)*#([^#\r\n].*)$/m');
array_push($replacements, "<li>$1</li>");
// second level
array_push($patterns, '/^((?:(?: |&nbsp;)*#[^\r\n][^\r\n]*[\r\n]{1,2})+)/ms');
array_push($replacements, "<li>\n<ol class=\"mediawiki\">\n$1</ol>\n</li>\n");
array_push($patterns, '/^(?: |&nbsp;)*##([^#\r\n].*)$/m');
array_push($replacements, "<li>$1</li>");
// third level
array_push($patterns, '/^((?:(?: |&nbsp;)*#[^\r\n][^\r\n]*[\r\n]{1,2})+)/ms');
array_push($replacements, "<li>\n<ol class=\"mediawiki\">\n$1</ol>\n</li>\n");
array_push($patterns, '/^(?: |&nbsp;)*###([^#\r\n].*)$/m');
array_push($replacements, "<li>$1</li>");
// fourth level
array_push($patterns, '/^((?:(?: |&nbsp;)*#[^\r\n][^\r\n]*[\r\n]{1,2})+)/ms');
array_push($replacements, "<li>\n<ol class=\"mediawiki\">\n$1</ol>\n</li>\n");
array_push($patterns, '/^(?: |&nbsp;)*####([^#\r\n].*)$/m');
array_push($replacements, "<li>$1</li>");

// : indented elements
array_push($patterns, '/^((?:(?: |&nbsp;)*:[^:\r\n][^\r\n]*[\r\n]{1,2})+)/ms');
array_push($replacements, "<dl class=\"mediawiki\">\n$1</dl>\n");
array_push($patterns, '/^(?: |&nbsp;)*:([^:\r\n].*)$/m');
array_push($replacements, "<dd>$1</dd>");
// second level
array_push($patterns, '/^((?:(?: |&nbsp;)*:[^\r\n][^\r\n]*[\r\n]{1,2})+)/ms');
array_push($replacements, "<dd>\n<dl class=\"mediawiki\">\n$1</dl>\n</dd>\n");
array_push($patterns, '/^(?: |&nbsp;)*::([^:\r\n].*)$/m');
array_push($replacements, "<dd>$1</dd>");
// third level
array_push($patterns, '/^((?:(?: |&nbsp;)*:[^\r\n][^\r\n]*[\r\n]{1,2})+)/ms');
array_push($replacements, "<dd>\n<dl class=\"mediawiki\">\n$1</dl>\n</dd>\n");
array_push($patterns, '/^(?: |&nbsp;)*:::([^:\r\n].*)$/m');
array_push($replacements, "<dd>$1</dd>");
// fourth level
array_push($patterns, '/^((?:(?: |&nbsp;)*:[^\r\n][^\r\n]*[\r\n]{1,2})+)/ms');
array_push($replacements, "<dd>\n<dl class=\"mediawiki\">\n$1</dl>\n</dd>\n");
array_push($patterns, '/^(?: |&nbsp;)*::::([^:\r\n].*)$/m');
array_push($replacements, "<dd>$1</dd>");

// tables
array_push($patterns, '/^(?: |&nbsp;)*\{\|(.*?)[\r\n](?:\|\+([^\r\n]*)[\r\n]+)?\|-[^|]*?\|(?:[^|\n\r]*\|)?(.*?)[\r\n]+\|\}/ms');
array_push($replacements, "<table width=\"100%\" cellspacing=\"1\" cellpadding=\"1\" border=\"1\" align=\"center\" class=\"mediawiki_table\">\n<caption class=\"mediawiki_table_legend\">$2</caption>\n<tbody>\n<tr>\n<td>$3\n&nbsp;</td>\n</tr>\n</tbody>\n</table>\n");
// <tr></tr>
array_push($patterns, '/^(?: |&nbsp;)*\|-[^|]*?\|(?:[^|\n\r]*\|)?/m');
array_push($replacements, "&nbsp;</td>\n</tr>\n<tr>\n<td>");
// <td></td>
array_push($patterns, '/^(?: |&nbsp;)*\|(?:[^|\n\r]*\|)?/m');
array_push($replacements, "&nbsp;</td>\n<td>");

// ----
array_push($patterns, '/----+/');
array_push($replacements, '<hr/>');

// {{Model}}
array_push($patterns, '/\{\{[^\}"\n\r]+\}\}/');
array_push($replacements, '');

// End of lines (not preceded by a HTML tag)
array_push($patterns, '/([^>] *)(?:\r\n|\n|\r)/s');
array_push($replacements, "$1<br/>\n");

$new_text = preg_replace($patterns, $replacements, $text);

return $new_text;

/**********************************/
/*** Mediawiki input filter END ***/
/**********************************/

}


/**
* convertToNumericHTMLEntities
*******************************
Convert the extended ASCII characters (for instance, characters with accents)
into valid HTML entities.

Parameters:
 $string (string): the content to convert. It can be UTF8-encoded.

Return:
 (string) the converted text.
*/
function convertToNumericHTMLEntities($string)
{
	// mapping for characters 0x80-0x9F
	$source_codes = array(
		chr(0x80),
		chr(0x81),
		chr(0x82),
		chr(0x83),
		chr(0x84),
		chr(0x85),
		chr(0x86),
		chr(0x87),
		chr(0x88),
		chr(0x89),
		chr(0x8A),
		chr(0x8B),
		chr(0x8C),
		chr(0x8D),
		chr(0x8E),
		chr(0x8F),
		chr(0x90),
		chr(0x91),
		chr(0x92),
		chr(0x93),
		chr(0x94),
		chr(0x95),
		chr(0x96),
		chr(0x97),
		chr(0x98),
		chr(0x99),
		chr(0x9A),
		chr(0x9B),
		chr(0x9C),
		chr(0x9D),
		chr(0x9E),
		chr(0x9F));		
	$remaped_codes = array(
		"&#8364;",
		" ",
		"&#8218;",
		"&#402;",
		"&#8222;",
		"&#8230;",
		"&#8224;",
		"&#8225;",
		"&#710;",
		"&#8240;",
		"&#352;",
		"&#8249;",
		"&#338;",
		" ",
		"&#381;",
		" ",
		" ",
		"&#8216;",
		"&#8217;",
		"&#8220;",
		"&#8221;",
		"&#8226;",
		"&#8211;",
		"&#8212;",
		"&#732;",
		"&#8482;",
		"&#353;",
		"&#8250;",
		"&#339;",
		" ",
		"&#382;",
		"&#376;");
	$string_ret = utf8_decode($string);
	// replace non alpha-numeric characters, non-space characters, non-apostrophe (0x27) and non 0x80-0x9F characters
	$string_ret = preg_replace(
		'/([\x00-\x08]|[\x0B-\x0C]|[\x0E-\x1F]|[\x7F]|[\xA0-\xFF])/e',
		'"&#".ord("$0").";"', $string_ret);
	// now remap 0x80-0x9F characters
	return str_replace($source_codes, $remaped_codes, $string_ret);
}


/**
* processMediaWikiXMLContents
******************************
Process the content of each XML block. Called by xml_parse().

*/
function processMediaWikiXMLContents($parser, $data)
{
    global $inside_page;
    global $inside_title;
    global $inside_text;
    global $inside_author;

    global $current_title;
    global $current_content;
    global $current_author;

    // check if we are inside a page definition
    if (TRUE == $inside_page)
    {
        // inside a page, check we goes the processed content
        if (TRUE == $inside_title)
        {
            // page title
            $current_title .= $data;
        }
        else if (TRUE == $inside_text)
        {
            // page text
            $current_content .= $data;
        }
        else if (TRUE == $inside_author)
        {
            // page contributor
            $current_author .= $data;
        }
    }
}


/**
* processMediaWikiXMLStartTag
******************************
Process each starting tag. Called by xml_parse().

*/
function processMediaWikiXMLStartTag($parser, $tag_name)
{
    global $pages_count;

    global $inside_page;
    global $inside_title;
    global $inside_text;
    global $inside_author;

    global $current_title;
    global $current_content;
    global $current_author;

    if ("PAGE" == $tag_name)
    {
        // <page> tag
        $inside_page = TRUE;
        // increase processed page counter
        ++$pages_count;
        // reset page data
        $current_title = '';
        $current_content = '';
        $current_author = '';
    }
    else if (TRUE == $inside_page)
    {
        if ("TITLE" == $tag_name)
        {
            $current_title = '';
            $inside_title = TRUE;
        }
        else if ("TEXT" == $tag_name)
        {
            $current_content = '';
            $inside_text = TRUE;
        }
        else if ("USERNAME" == $tag_name)
        {
            $current_author = '';
            $inside_author = TRUE;
        }
        else
        {
            // not one of the above tags
            $inside_title = FALSE;
            $inside_text = FALSE;
            $inside_author = FALSE;
        }
    }
}


/**
* processMediaWikiXMLEndTag
****************************
Process each ending tag. Called by xml_parse().

*/
function processMediaWikiXMLEndTag($parser, $tag_name)
{
    global $imported_pages_count;

    global $inside_page;
    global $inside_title;
    global $inside_text;
    global $inside_author;

    global $current_title;
    global $current_content;
    global $current_author;

    if ("PAGE" == $tag_name)
    {
        // done with current page, save it
        $inside_page = FALSE;
        drupal_set_message("Creating page " . ++$imported_pages_count . ": <i>" . htmlentities(utf8_decode($current_title)) . "</i>\n", 'status');
        $new_node = (object) array();
        $new_node->nid = $new_node->vid = NULL;
        $new_node->title = substr($current_title, 0, 255);
        $new_node->status = 1; // published
        $new_node->promote = 0;
        $new_node->sticky = 0;
        $new_node->format = $GLOBALS['mediawiki_input_format_number']; // Input Format
        $current_content = html_entity_decode($current_content); // decode what's encoded in XML file
        $current_content = convertToNumericHTMLEntities($current_content); // MediaWiki export may leave invalid charcters! We have to replace them.
        if (TRUE == $GLOBALS['convert_mediawiki_tags_to_html'])
        {
            $current_content = convertMediaWikiToHTML($current_content);
        }
        $new_node->body = $current_content;
        $new_node->teaser = '';
        $new_node->type = $GLOBALS['node_type'];
        // author: if not in the lookup table, try name as is and otherwise "Anonymous" is used
        if (TRUE == $GLOBALS['import_user_names'])
        {
            if ((array_key_exists($current_author, $GLOBALS['user_name_lookup']))
                && $contributor = user_load(array('name' => $GLOBALS['user_name_lookup'][$current_author])))
            {
                $new_node->uid = $contributor->uid;
            }
            else if ($contributor = user_load(array('name' => $current_author)))
            {
                $new_node->uid = $contributor->uid;
            }
            /*
            else if ('' != $current_author)
            {
                drupal_set_message("No valid Drupal name found for author '$current_author'\n", 'warning');
            }
            */
        }
        else
        {
            global $user;
            $new_node->uid = $user->uid;
        }
        // record page
        node_save($new_node);
        if (empty($new_node->nid))
        {
            --$imported_pages_count;
            drupal_set_message("Page import failed!\n", 'status');
        }

        // reset data
        $current_title = '';
        $current_content = '';
        $current_author = '';
    }
    else if (TRUE == $inside_page)
    {
        // process other closing tags
        if ("TITLE" == $tag_name)
        {
            $inside_title = FALSE;
        }
        else if ("TEXT" == $tag_name)
        {
            $inside_text = FALSE;
        }
        else if ("USERNAME" == $tag_name)
        {
            $inside_author = FALSE;
        }
    }
}

/**
* processMediaWikiXML
**********************
Process Media Wiki XML content.

Parameters:
 $xml_data (string): XML content to process.

Return:
 (integer) an error code of an error occured, 0 otherwise.
*/
function processMediaWikiXML($xml_data)
{
    $error_code = 0;
    $xml_parser = xml_parser_create();

    xml_set_element_handler($xml_parser, "processMediaWikiXMLStartTag", "processMediaWikiXMLEndTag");

    xml_set_character_data_handler($xml_parser, "processMediaWikiXMLContents");

    if (!xml_parse($xml_parser, $xml_data, TRUE))
    {
        // report the error
        drupal_set_message('XML parsing error on line ' . xml_get_current_line_number($xml_parser), 'error');
        $error_code = 1;
    }

    xml_parser_free($xml_parser);

    return $error_code;
}


/**
* mediawiki_import_form
************************
Importation form data. Called by Drupal.

*/
function mediawiki_import_form($form_state)
{
    $form['#attributes'] = array('enctype' => "multipart/form-data");

    $form['xml_file'] = array(
            '#type' => 'file',
            '#title' => t('XML file to import'),
            '#size' => 30,
            '#description' => t('Import an XML MediaWiki file.'),
        );
/*
//+FIXME: other fields that could be usefull and could be implemented:

    $form['files_url'] = array(
            '#type' => 'textfield',
            '#title' => t('URL to MediaWiki files (including the ending "/")'),
            '#default_value' => url('', array('absolute' => TRUE)) . 'sites/default/files/mediawiki/',
            '#size' => 60,
            '#maxlength' => 256,
            '#description' => t('Enter the URL to media and image files with the ending "/".'),
        );

    $form['node_type'] = array(
            '#type' => 'textfield',
            '#title' => t('Node type'),
            '#size' => 30,
            '#maxlength' => 64,
            '#description' => t('Enter the node type to use for imported MediaWiki pages. The node type must be an existing type.'),
        );

    $form['authors'] = array(
            '#type' => 'textarea',
            '#title' => t('Lookup table for authors'),
            '#default_value' =>  'wiki_name=drupal_name',
            '#cols' => 60,
            '#rows' => 5,
            '#description' => t('Enter wich MediaWiki contributor name corresponds to wich Drupal user name using the format: "wiki_name=drupal_name" (one per line). '),
        );
*/
    $form['submit'] = array('#type' => 'submit', '#value' => t('Import'));

    return $form;
}

/**
* mediawiki_import_form_submit
*******************************
Process form submission. Called by Drupal.

*/
function mediawiki_import_form_submit($form, &$form_state)
{
    global $pages_count;
    global $imported_pages_count;

    $validators = array();
    // save temporary XML file
    $target = file_directory_temp();
    $file = file_save_upload('xml_file', $validators, $target);

    if ($file != 0)
    {
        // file exists, process XML...
        if (0 == processMediaWikiXML(file_get_contents($file->filepath)))
        {
            // everything OK
            drupal_set_message("MediaWiki XML imporation: OK; $imported_pages_count pages imported over $pages_count processed.", 'status');
        }
        else
        {
            // failure
            drupal_set_message('MediaWiki XML imporation: an error occured!', 'error');
        }
    }
    else
    {
        // unble to create file (invalid directory? access rights?) or empty file
        form_set_error('myform', t("Failed to save the Media Wiki XML file to import!"));
        drupal_set_message('MediaWiki XML imporation: an error occured!', 'error');
    }
}

/*****************************************************************************/
// CODE START

// display form or process import
echo drupal_get_form('mediawiki_import_form');

/*****************************************************************************/

?>
